from __future__ import annotations

from joblib_windower.ndarray_windower import ndarray_windower
from joblib_windower.ndframe_windower import ndframe_windower
from joblib_windower.slide_ndarrays import slide_ndarrays
from joblib_windower.slide_ndframes import slide_ndframes


__all__ = [
    "ndarray_windower",
    "ndframe_windower",
    "slide_ndarrays",
    "slide_ndframes",
]
__version__ = "0.3.10"



from __future__ import annotations

from functools import partial
from functools import wraps
from itertools import chain
from multiprocessing import cpu_count
from pathlib import Path
from tempfile import gettempdir
from tempfile import TemporaryDirectory
from typing import Any
from typing import Callable
from typing import Optional
from typing import TypeVar
from typing import Union

import joblib
from atomic_write_path import atomic_write_path
from functional_itertools import CDict
from functional_itertools import CList
from joblib import delayed
from joblib import Parallel
from numpy import array
from numpy import memmap
from numpy import nan
from numpy import ndarray

from joblib_windower.errors import InvalidWindowError


CPU_COUNT = cpu_count()
TEMP_DIR = gettempdir()
T = TypeVar("T")


def _writes_output(func: Callable[..., Any]) -> Callable[..., None]:
    @wraps(func)
    def wrapped(*args: Any, _output: ndarray, _index: int, **kwargs: Any) -> None:
        _output[_index] = func(*args, **kwargs)

    return wrapped


def _applies_slice(func: Callable[..., T]) -> Callable[..., T]:
    @wraps(func)
    def wrapped(*args: Any, _indexer: Union[int, slice], **kwargs: Any) -> T:
        applier = partial(_maybe_slice, indexer=_indexer)
        return func(*CList(args).map(applier), **CDict(kwargs).map_values(applier))

    return wrapped


def _maybe_slice(x: Any, *, indexer: Union[int, slice]) -> Any:
    if isinstance(x, memmap):
        return x[indexer]
    else:
        return x


def _maybe_memmap(x: Any, path: Union[Path, str]) -> Any:
    if isinstance(x, ndarray):
        with atomic_write_path(path) as temp:
            joblib.dump(x, temp)
        return joblib.load(path, mmap_mode="r")
    else:
        return x


def _build_array_windower(
    func: Optional[Callable[..., Union[float, ndarray]]] = None,
    *,
    temp_dir: Union[Path, str] = TEMP_DIR,
) -> Callable[..., ndarray]:
    @wraps(func)
    def wrapped(
        *args: Any,
        window: int = 1,
        min_frac: Optional[float] = None,
        n_jobs: int = CPU_COUNT,
        **kwargs: Any,
    ) -> ndarray:
        if window <= 0:
            raise InvalidWindowError(f"Got window = {window}")

        applies_slice = _applies_slice(func)
        Path(temp_dir).mkdir(parents=True, exist_ok=True)
        with TemporaryDirectory(dir=temp_dir) as td:
            td = Path(td)

            # maybe memmap
            new_args: CList[Any] = (
                CList(args)
                .enumerate()
                .starmap(lambda i, v: _maybe_memmap(v, td.joinpath(f"arg_{i}")))
            )
            new_kwargs: CDict[str, Any] = CDict(kwargs).map_items(
                lambda k, v: (k, _maybe_memmap(v, td.joinpath(f"kwarg_{k}"))),
            )
            length = (
                new_args.chain(new_kwargs.values())
                .filter(lambda x: isinstance(x, memmap))
                .map(len)
                .set()
                .one()
            )
            if length == 0:
                raise ValueError("Expected non-zero length")

            # create indexers
            if window == 1:
                indexers: CList[int] = CList.range(length)
            else:
                indexers: CList[Optional[slice]] = CList.range(1, length + 1).map(
                    lambda x: slice(max(x - window, 0), x),
                )
                if min_frac is None:
                    indexers: CList[Optional[slice]] = indexers
                else:
                    min_length = min_frac * window
                    indexers = indexers.map(
                        lambda x: x if (x.stop - x.start) >= min_length else None,
                    )

            # apply last index
            *_, last_indexer = indexers
            if last_indexer is None:
                raise ValueError("Expected the last indexer to be a slice")
            last_result = applies_slice(*new_args, _indexer=last_indexer, **new_kwargs)
            last_array = array(last_result)

            # prepare output
            output = memmap(
                filename=str(td.joinpath("output")),
                dtype=last_array.dtype,
                mode="w+",
                shape=tuple(chain([length], last_array.shape)),
            )

            # apply all remaining slices
            Parallel(n_jobs=n_jobs)(
                delayed(_writes_output(applies_slice))(
                    *new_args, _indexer=indexer, _output=output, _index=index, **new_kwargs,
                )
                for index, indexer in enumerate(indexers)
                if indexer not in [None, last_indexer]
            )

        # NaN head and copy last result
        out_array = array(output)
        for i, _ in indexers.enumerate().filter(lambda x: x[1] is None):
            out_array[i] = nan
        out_array[-1] = last_result
        return out_array

    return wrapped


def ndarray_windower(
    func: Optional[Callable[..., Union[float, ndarray]]] = None,
    *,
    temp_dir: Union[Path, str] = TEMP_DIR,
) -> Callable[..., ndarray]:
    if func is None:
        return partial(ndarray_windower, temp_dir=temp_dir)
    else:
        return _build_array_windower(func, temp_dir=temp_dir)



from __future__ import annotations

from functools import partial
from functools import wraps
from operator import attrgetter
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Dict
from typing import Hashable
from typing import Iterable
from typing import Optional
from typing import Tuple
from typing import Union

from functional_itertools import CDict
from functional_itertools import CList
from numpy import ndarray
from pandas import DataFrame
from pandas import Index
from pandas import Series
from pandas.testing import assert_index_equal

from joblib_windower import ndarray_windower
from joblib_windower.errors import InvalidWindowError
from joblib_windower.ndarray_windower import CPU_COUNT
from joblib_windower.ndarray_windower import TEMP_DIR


def to_numpy(x: Union[Index, Series, DataFrame]) -> ndarray:
    return x.to_numpy()


def _index_to_numpy(x: Union[Series, DataFrame]) -> ndarray:
    array = to_numpy(x.index)
    try:
        return array.astype(str)
    except ValueError:
        return array


def _maybe_to_numpy(x: Any) -> Tuple[Any, Optional[ndarray], Optional[Index]]:
    if isinstance(x, Series):
        return to_numpy(x), _index_to_numpy(x), None
    elif isinstance(x, DataFrame):
        return to_numpy(x), _index_to_numpy(x), x.columns
    else:
        return x, None, None


def _maybe_to_pandas(value: Any, index: Any, columns: Optional[Index]) -> Any:
    if isinstance(value, ndarray):
        if index is None:
            raise ValueError("Expected ndarrays to come with indices")
        elif isinstance(index, ndarray):
            if columns is None:
                return Series(value, index=index)
            else:
                return DataFrame(value, index=index, columns=columns)
        else:
            if columns is None:
                raise NotImplementedError("Unexpected case")
            else:
                return Series(value, index=columns, name=index)
    else:
        return value


def _build_internal(*, temp_dir: Union[Path, str] = TEMP_DIR) -> Callable[..., ndarray]:
    @ndarray_windower(temp_dir=temp_dir)
    def internal(
        *args: Tuple[Any, Optional[Index], Optional[Index]],
        _func: Callable[..., Union[float, Series]],
        _maybe_columns_args: Tuple[Optional[Index], ...],
        _maybe_columns_kwargs: Dict[str, Optional[Index]],
        **kwargs: Tuple[Any, Optional[Index], Optional[Index]],
    ) -> Any:
        # reassemble arguments
        new_args: CList[Any] = CList()
        args, _maybe_columns_args = CList(args), CList(_maybe_columns_args)
        while args:
            new_args.append(
                _maybe_to_pandas(
                    value=args.pop(0), index=args.pop(0), columns=_maybe_columns_args.pop(0),
                ),
            )

        new_kwargs: CDict[str, Any] = CDict()
        while kwargs:
            key = next(iter(kwargs))
            try:
                value, index = kwargs.pop(key), kwargs.pop(f"_{key}")
            except KeyError:
                key = key[1:]
                value, index = kwargs.pop(key), kwargs.pop(f"_{key}")
            new_kwargs[key] = _maybe_to_pandas(
                value=value, index=index, columns=_maybe_columns_kwargs[key],
            )

        return _func(*new_args, **new_kwargs)

    return internal


def _build_ndframe_windower(
    func: Optional[Callable[..., Union[float, Series]]] = None,
    *,
    temp_dir: Union[Path, str] = TEMP_DIR,
) -> Callable[..., Union[Series, DataFrame]]:
    @wraps(func)
    def wrapped(
        *args: Any,
        window: int = 1,
        min_frac: Optional[float] = None,
        n_jobs: int = CPU_COUNT,
        columns: Optional[Iterable[Hashable]] = None,
        **kwargs: Any,
    ) -> Union[Series, DataFrame]:
        args, kwargs = CList(args), CDict(kwargs)

        # check for existence of and get unique index
        indices = (
            args.chain(kwargs.values())
            .filter(lambda x: isinstance(x, (Series, DataFrame)))
            .map(attrgetter("index"))
        )
        if not indices:
            raise ValueError("Expected at least 1 Series or DataFrame; got none")
        for index1, index2 in indices.combinations(2):
            assert_index_equal(index1, index2)
        index, *_ = indices
        if window <= 0:
            raise InvalidWindowError(f"Got window = {window}")

        # maybe decompose arguments
        try:
            maybe_numpy_args, maybe_index_args, maybe_columns_args = args.map(
                _maybe_to_numpy,
            ).unzip()
        except ValueError:
            maybe_numpy_args = maybe_index_args = maybe_columns_args = CList()
        try:
            maybe_numpy_kwargs, maybe_index_kwargs, maybe_columns_kwargs = (
                kwargs.map_values(_maybe_to_numpy).values().unzip()
            )
            maybe_numpy_kwargs = kwargs.keys().zip(maybe_numpy_kwargs).dict()
            maybe_index_kwargs = kwargs.keys().zip(maybe_index_kwargs).dict()
            maybe_columns_kwargs = kwargs.keys().zip(maybe_columns_kwargs).dict()
        except ValueError:
            maybe_numpy_kwargs = maybe_index_kwargs = maybe_columns_kwargs = CDict()

        # pass decomposed arguments to internal function
        result: ndarray = _build_internal(temp_dir=temp_dir)(
            *maybe_numpy_args.zip(maybe_index_args).flatten(),
            _func=func,
            _maybe_columns_args=maybe_columns_args,
            _maybe_columns_kwargs=maybe_columns_kwargs,
            window=window,
            min_frac=min_frac,
            n_jobs=n_jobs,
            **maybe_numpy_kwargs,
            **maybe_index_kwargs.map_keys(lambda x: f"_{x}"),
        )

        # reassemble result
        if result.ndim == 1:
            return Series(result, index=index)
        elif result.ndim == 2:
            if columns is None:
                df_columns = (
                    args.chain(kwargs.values())
                    .filter(lambda x: isinstance(x, DataFrame))
                    .map(attrgetter("columns"))
                )
                if df_columns:
                    try:
                        for columns1, columns2 in df_columns.combinations(2):
                            assert_index_equal(columns1, columns2)
                    except AssertionError:
                        df_columns_use = None
                    else:
                        df_columns_use, *_ = df_columns
                else:
                    df_columns_use = None
            else:
                df_columns_use = columns
            return DataFrame(result, index=index, columns=df_columns_use)
        else:
            raise ValueError(f"Expected 1 or 2 dimensions; got {result.ndim}")

    return wrapped


def ndframe_windower(
    func: Optional[Callable[..., Union[float, ndarray]]] = None,
    *,
    temp_dir: Union[Path, str] = TEMP_DIR,
) -> Callable[..., Union[Series, DataFrame]]:
    if func is None:
        return partial(ndframe_windower, temp_dir=temp_dir)
    else:
        return _build_ndframe_windower(func, temp_dir=temp_dir)



from __future__ import annotations


class InvalidDTypeError(ValueError):
    """Raised when the dtype is invalid."""


class InvalidMinFracError(ValueError):
    """Raised when the minimum fraction is invalid."""


class InvalidLengthError(ValueError):
    """Raised when the length is invalid."""


class InvalidStepError(ValueError):
    """Raised when the step is invalid."""


class InvalidWindowError(ValueError):
    """Raised when the window is invalid."""


class NoSlicersError(ValueError):
    """Raised when no slicers are valid."""



from __future__ import annotations

from functools import partial
from operator import itemgetter
from pathlib import Path
from typing import Any
from typing import Callable
from typing import Hashable
from typing import Optional
from typing import Tuple
from typing import TypeVar
from typing import Union

from attr import attrs
from functional_itertools import CIterable
from functional_itertools import CList
from functional_itertools import CTuple
from functional_itertools import EmptyIterableError
from functional_itertools import MultipleElementsError
from numpy import bool_
from numpy import dtype
from numpy import issubdtype
from numpy import nan
from numpy import ndarray
from numpy import number
from numpy import str_
from numpy.ma import MaskedArray
from pandas import concat
from pandas import DataFrame
from pandas import Index
from pandas import Series
from pandas.testing import assert_index_equal

from joblib_windower.slide_ndarrays import slide_ndarrays
from joblib_windower.utilities import Arguments
from joblib_windower.utilities import CPU_COUNT
from joblib_windower.utilities import datetime64ns
from joblib_windower.utilities import DEFAULT_STR_LEN_FACTOR
from joblib_windower.utilities import is_not_none
from joblib_windower.utilities import NaT
from joblib_windower.utilities import pandas_obj_to_ndarray
from joblib_windower.utilities import TEMP_DIR
from joblib_windower.utilities import timedelta64ns

T = TypeVar("T")


@attrs(auto_attribs=True)
class DataFrameMetadata:
    index_name: Hashable
    columns: Index


@attrs(auto_attribs=True)
class IndexMetadata:
    name: Hashable


@attrs(auto_attribs=True)
class NDFrameSpec:
    dtype: dtype
    masked: Any


@attrs(auto_attribs=True)
class SeriesMetadata:
    name: Hashable
    index_name: Hashable


def call_with_packing(keys: CTuple[Optional[str]], func: Callable[..., T], *args: Any) -> T:
    packed = CTuple(args).grouper(3).map(pack_argument)
    is_kwargs, is_args = keys.zip(packed).partition(lambda x: x[0] is None)
    arguments = Arguments(args=is_args.map(itemgetter(1)), kwargs=is_kwargs.dict())
    return func(*arguments.args, **arguments.kwargs)


def get_maybe_dataframe_columns(x: Any) -> Optional[Index]:
    if isinstance(x, DataFrame):
        return x.columns
    else:
        return None


def get_maybe_ndframe_index(x: Any) -> Optional[Index]:
    if isinstance(x, (Series, DataFrame)):
        return x.index
    else:
        return None


def get_maybe_series_name(value: Any) -> Optional[Hashable]:
    if isinstance(value, Series):
        return value.name
    else:
        return None


def get_maybe_unique_dataframe_columns(arguments: Arguments) -> Optional[Index]:
    columns = (
        arguments.map_values(get_maybe_dataframe_columns)
        .all_values()
        .filter(is_not_none)
        .map(Index)
    )
    if columns:
        try:
            for columns1, columns2 in columns.combinations(2):
                assert_index_equal(columns1, columns2)
        except AssertionError:
            return None
        else:
            first_columns, *_ = columns
            return first_columns
    else:
        return None


def get_maybe_unique_series_name(arguments: Arguments) -> Optional[Hashable]:
    names = arguments.map_values(get_maybe_series_name).all_values().filter(is_not_none).set()
    try:
        return names.one()
    except (EmptyIterableError, MultipleElementsError):
        return None


def get_maybe_unique_ndframe_index(arguments: Arguments) -> Optional[Index]:
    indices = arguments.map_values(get_maybe_ndframe_index).all_values().filter(is_not_none)
    if indices:
        try:
            for index1, index2 in indices.combinations(2):
                assert_index_equal(index1, index2)
        except AssertionError:
            return None
        else:
            index, *_ = indices
            return index
    else:
        raise ValueError("Expected at least 1 Series or DataFrame; got none")


def get_ndframe_spec(x: dtype) -> NDFrameSpec:
    if CIterable([bool_, str_]).map(lambda y: issubdtype(x, y)).any():
        return NDFrameSpec(dtype=dtype(object), masked=nan)
    elif issubdtype(x, number):
        return NDFrameSpec(dtype=dtype(float), masked=nan)
    elif x in {datetime64ns, timedelta64ns}:
        return NDFrameSpec(dtype=datetime64ns, masked=NaT)
    else:
        raise ValueError(f"Unable to convert dtype {x} to an NDFrameSpec")


def masked_array_to_pandas_object(
    array: MaskedArray, index: Index, name: Optional[Hashable], columns: Optional[Index],
) -> Union[Series, DataFrame]:
    spec = get_ndframe_spec(array.dtype)
    if array.ndim == 1:
        return Series(data=array.data, index=index, dtype=spec.dtype, name=name).where(
            ~array.mask, spec.masked,
        )
    elif array.ndim == 2:
        m, n = array.shape
        return DataFrame(
            data=array.data,
            index=index if len(index) == m else None,
            columns=columns if columns is not None and len(columns) == n else None,
        ).where(~array.mask, spec.masked)
    elif array.ndim == 3:
        length, *_ = array.shape
        dfs = CList.range(length).map(
            lambda x: masked_array_to_pandas_object(
                array=array[x], index=index, name=name, columns=columns,
            ),
        )
        return concat(dfs, axis=0, keys=index)
    else:
        raise ValueError(f"Expected 1-3 dimensions; got {array.ndim}")


def pack_argument(
    data: Tuple[
        Optional[Union[IndexMetadata, SeriesMetadata, DataFrameMetadata]], Any, Optional[ndarray],
    ],
) -> Any:
    metadata, x, maybe_index = data
    if isinstance(metadata, IndexMetadata):
        return Index(x, name=metadata.name)
    elif isinstance(metadata, SeriesMetadata):
        if isinstance(maybe_index, ndarray):
            return Series(
                x, index=Index(maybe_index, name=metadata.index_name), name=metadata.name,
            )
        else:
            return x
    elif isinstance(metadata, DataFrameMetadata):
        if isinstance(maybe_index, ndarray):
            return DataFrame(
                x, index=Index(maybe_index, name=metadata.index_name), columns=metadata.columns,
            )
        else:
            return Series(x, index=metadata.columns, name=maybe_index)
    else:
        return x


def slide_ndframes(
    func: Callable,
    *args: Any,
    window: int = 1,
    min_frac: Optional[float] = None,
    step: int = 1,
    temp_dir: Union[Path, str] = TEMP_DIR,
    str_len_factor: int = DEFAULT_STR_LEN_FACTOR,
    parallel: bool = False,
    processes: int = CPU_COUNT,
    **kwargs: Any,
) -> Union[Series, DataFrame]:
    arguments = Arguments(args=args, kwargs=kwargs)
    index = get_maybe_unique_ndframe_index(arguments)
    maybe_name = get_maybe_unique_series_name(arguments)
    maybe_columns = get_maybe_unique_dataframe_columns(arguments)
    unpacked_arguments = arguments.all_values().map(unpack_argument)
    keys = arguments.args.map(lambda _: None).chain(arguments.kwargs.keys())
    masked_array = slide_ndarrays(
        partial(call_with_packing, keys, func),
        *unpacked_arguments.flatten(),
        window=window,
        min_frac=min_frac,
        step=step,
        temp_dir=temp_dir,
        str_len_factor=str_len_factor,
        parallel=parallel,
        processes=processes,
    )
    return masked_array_to_pandas_object(masked_array, index, maybe_name, maybe_columns)


def unpack_argument(x: Any) -> Tuple[Any, Any, Optional[ndarray]]:
    if isinstance(x, Index):
        array, name = unpack_index(x)
        return IndexMetadata(name=name), array, None
    elif isinstance(x, Series):
        return unpack_series(x)
    elif isinstance(x, DataFrame):
        return unpack_dataframe(x)
    else:
        return None, x, None


def unpack_dataframe(x: DataFrame) -> Tuple[DataFrameMetadata, ndarray, ndarray]:
    return (
        DataFrameMetadata(index_name=x.index.name, columns=x.columns),
        pandas_obj_to_ndarray(x),
        pandas_obj_to_ndarray(x.index),
    )


def unpack_index(x: Index) -> Tuple[ndarray, Hashable]:
    return pandas_obj_to_ndarray(x), x.name


def unpack_series(x: Series) -> Tuple[SeriesMetadata, ndarray, ndarray]:
    return (
        SeriesMetadata(name=x.name, index_name=x.index.name),
        pandas_obj_to_ndarray(x),
        pandas_obj_to_ndarray(x.index),
    )
